// -*- C++ -*-

#ifndef _GCDIEHARD_H_
#define _GCDIEHARD_H_

#include <assert.h>

#include "../diehardheap.h"
#include "gccfg.h"

struct FormattedSize {
	char	mark;
	float	size;
};

template <int Numerator,
	  int Denominator,
	  int MaxSize,
	  bool DieFast>

class GCDieHardHeap : public DieHardHeap<Numerator, Denominator, MaxSize, DieFast> {

	typedef DieHardHeap<Numerator, Denominator, MaxSize, DieFast> ParentHeap;
	
	using ParentHeap::MAX_INDEX;
	using ParentHeap::getHeap;
	using ParentHeap::getIndex;
	using ParentHeap::getClassSize;

public:
	
	GCDieHardHeap (void)
		: _allocs_this_session (0),
			_bytes_allocated (0),
			_liveSize (0),
			_heapSize (0),
			_total_alloc (0),
			_total_collections (0)
		{ }
		
	inline void * malloc (size_t sz) {
		int index = getIndex (sz);
		int mh = getHeap(index)->objectsAvailable();
		
		void * ptr = ParentHeap::malloc (sz);
		
		// A lot of work for tracking allocation and heap growth
		if (ptr != NULL) {
			_allocs_this_session++;
			_bytes_allocated += ParentHeap::getSize (ptr);

			int cs = getClassSize (index);
			_liveSize += cs;
			
			if (getHeap(index)->objectsAvailable() > mh) {
				_heapSize = 0;
				for (int i = 0; i < MAX_INDEX; i++) {
  				_heapSize += (getHeap(i)->objectsAvailable() * getClassSize (i));
  			}
  		}
		}

		return ptr;
	}
	
	/// Check all the randomheaps for the given pointer, mark it if found
  inline int markScan (void * ptr) {
  	for (int i = 0; i < MAX_INDEX; i++) {
  		int mb = getHeap(i)->markScan (ptr);
  		if (mb != 0)
  			return mb;
  		}
  	
  	// Object could be "big"
  	return 0;
  }
  
  inline void fullScan (void) {
  	for (int i = 0; i < MAX_INDEX; i++) {
  		getHeap(i)->fullScan ();
  	}
  }
	
	/// Not a sweep in the GC sense, but "sweeps" all the miniheap bitmaps 
  /// squeaky clean.  Also gets the live size prior to sweeping.
  inline void gcZeroBitmaps (void) {
  	if (GC_VERBOSE && GC_VERBOSE_TRACE) {
  		char buf[255];
  		sprintf (buf, "*** ZERO out bitmaps\n");
  		fprintf (stderr, buf);
  	}

    for (int i = 0; i < MAX_INDEX; i++) {
      getHeap(i)->zeroBitmap();
    }
    
    _liveSize = 0;
  }
		
	/// When USING_FREE is enabled, checks all heaps for inconsistencies.
  inline void gcBitmapCheck (void) {
  	if (GC_VERBOSE && GC_VERBOSE_BITMAPS) {
	  	char buf[255];
 	 		sprintf (buf, "*** Starting GC/free comparison\n");
 	 		fprintf (stderr, buf);
 	 	}
  	
  	for (int i = 0; i < MAX_INDEX; i++) {
  		getHeap(i)->bitmapCheck();
  	}
  }
  
  /// Forces a default value into all the dead memory
  inline void gcBuryDead (void) {
  	if (GC_VERBOSE && GC_VERBOSE_TRACE) {
  		char buf[255];
  		sprintf (buf, "*** BURY dead data\n");
  		fprintf (stderr, buf);
  	}
  	
  	for (int i = 0; i < MAX_INDEX; i++) {
  		getHeap(i)->buryDead();
  	}
  }
  
  inline void * gcResolve (size_t gcclass, size_t gcheap, size_t gcindex) {
  	int idx = getIndex (gcclass);
  	return getHeap(idx)->resolve (gcheap, gcindex);
  }
  
  inline int gcLiveSize (void) {
  	return _liveSize;
  }
  
  inline int gcHeapSize (void) {
  	return _heapSize;
  }
  
  /// Attempts to grow every randomheap, given the current Root Size and Live
  /// Size.  Also prints a plethora of statistical information.
  inline void gcGrow (size_t rootSize, size_t allocated = 0) {
  	int liveSize = 0;
  	int heapSize = 0;
  	
  	_total_collections++;
  	
  	if (GC_VERBOSE && GC_VERBOSE_TRACE) {
  		char buf[255];
  		sprintf (buf, "*** Perform GROWTH check\n");
  		fprintf (stderr, buf);
  	}
  	
  	// Get the current live and heap sizes
  	for (int i = 0; i < MAX_INDEX; i++) {
  		liveSize += (getHeap(i)->objectsInUse() * getClassSize (i));
  	}
  	
  	for (int i = 0; i < MAX_INDEX; i++) {
  		heapSize += (getHeap(i)->objectsAvailable() * getClassSize (i));
  	}
  	
  	_total_alloc += (allocated / 1024.0);

  	if (GC_VERBOSE && GC_VERBOSE_GROW) {
  		char buf[255];
  		sprintf (buf, "Live Size: %i, Root Size: %i\n",  liveSize, rootSize);
  		fprintf (stderr, buf);
  	}
  	
  	// Calculate the GC Workload, mark/cons, for the period since the last
  	// allocation.
  	float per_time = ((liveSize + rootSize) * 1.0) / (liveSize + rootSize + _allocs_this_session);
  	_allocs_this_session = 0;
  	
  	float real_time = ((liveSize + rootSize) * 1.0) / (liveSize + rootSize + allocated);
  	_bytes_allocated = 0;
  	
  	if (GC_VERBOSE && GC_VERBOSE_GROW) {
	  	char sbuf[255];
 	 		sprintf (sbuf, "[%i] GC Time: %.6f (%.6f), Heap Size: %i KB, Total Alloc: %.2f MB\n", _total_collections, per_time, real_time, (heapSize / 1024), (_total_alloc / 1024.0) );
  		fprintf (stderr, sbuf);
  	}
  	
  	// Try growing the randomheaps, given K + L in the invariant
  	// Hi/C > K + L + Li
  	for (int i = 0; i < MAX_INDEX; i++) {
  		getHeap(i)->gcGrow(rootSize + liveSize);
  	}

    // Calculate and print the new memory footprint for the application.
    int newLiveSize = 0;
    for (int i = 0; i < MAX_INDEX; i++) {
  		newLiveSize += (getHeap(i)->objectsAvailable() * getClassSize (i));
  	}
  	
  	_liveSize = liveSize;
  	_heapSize = newLiveSize;
  	
  	FormattedSize fs = formatSize (newLiveSize);
		
		if (GC_VERBOSE && GC_VERBOSE_GROW) {
			char sbuf[255];
  		sprintf (sbuf, "Current Memory Footprint: %.2f %cB\n", fs.size, fs.mark);
  		fprintf (stderr, sbuf);
  	}
  }
  
  /// Pretty print
  inline FormattedSize formatSize (int sz) {
  	FormattedSize fs;
  	
  	if (sz >= 1024*1024*1024) {
  		fs.mark = 'G';
  		fs.size = (sz * 1.0) / (1024*1024*1024);
  	} else if (sz >= 1024*1024) {
  		fs.mark = 'M';
  		fs.size = (sz * 1.0) / (1024*1024);
  	} else if (sz >= 1024) {
  		fs.mark = 'K';
  		fs.size = (sz * 1.0) / 1024;
  	} else {
  		fs.mark = ' ';
  		fs.size = sz;
  	}
  	
  	return fs;
  }
  
  /// Counts all the live objects in all the randomheaps
  inline int objectsInUse (void) {
  	int cnt = 0;
  	for (int i = 0; i < MAX_INDEX; i++) {
      cnt += getHeap(i)->objectsInUse();
    }
    return cnt;
  }
  
private:
	
	int _allocs_this_session;
	int _bytes_allocated;
	int _total_collections;
	double _total_alloc;
	
	int _liveSize;
  int _heapSize;
};

#endif
