// -*- C++ -*-

#ifndef _GC_DYNAMIC_
#define _GC_DYNAMIC_

#include <assert.h>
#include <link.h>

#include "gccfg.h"
#include "dynarray.h"
#include "gcplatform.h"
#include "../mmapwrapper.h"

#pragma weak dl_iterate_phdr

/// Small object to identify a region of memory as a root.
class GC_root {
public:
	void * start;
	void * end;
	bool tmp;
};

class GCRootSet {
public:
	
	GCRootSet (void)
		:	_n_roots (0),
			_root_size (0)
		{ }
	
	/// Check if a root at the given pointer is already registered
	GC_root * rootsPresent (char * b) {
	
    // int h = rt_hash(b);
    // GC_root * p = GC_root_index[h];
    
    // Boehm solves this with a hashing function.  At the moment we're taking
    // the silly approach of a sequential search.  We'll probably want to 
    // change that.
    
    //while (p != 0) {
    //    if (p -> r_start == (ptr_t)b) return(p);
    //    p = p -> r_next;
    //}
    
    for (int i=0; i < _n_roots; i++) {
    	if ((char *) _rootset[i].start == b) {
    		return &_rootset[i];
    	}
    }

    return 0;
	}

	/// Add a region of memory to the root set.  If it's marked as temporary, it
	/// will be deleted before the next collection.
	void addRoots (char * b, char * e, bool tmp) {
		GC_root * old = rootsPresent (b);
		
    if (old != 0) {
      if (e <= (char *) old->end) {
      	// Root already exists
      	return;
      }
      
      // Root needs to be extended
      _root_size += e - (char *) old->end;
      old->end = (void *) e;
      return;
    }
    
    _rootset[_n_roots].start = (void *) b;
    _rootset[_n_roots].end = (void *) e;
    _rootset[_n_roots].tmp = tmp;
    // _rootset[_n_roots].r_next = 0;
    
    //add_roots_to_index(GC_static_roots + n_root_sets);
    _root_size += e - b;
    _n_roots++;
	}
	
	/// Delete any roots contained in the given memory range
	inline void deleteRoots (char * b, char * e) {
		for (int i=0; i < _n_roots;) {
    	if (_rootset[i].start >= (char *) b && _rootset[i].end <= (char *) e) {
    		deleteRootsAtPos (i);
    	} else {
    		i++;
    	}
    }
	}
	
	/// Delete all temporary roots from the set
	inline void deleteTempRoots (void) {
  	for (int i=0; i < _n_roots;) {
    	if (_rootset[i].tmp) {
        deleteRootsAtPos (i);
    	} else {
    		i++;
    	}
    }
  }
	
	/// Delete a given root from the root set
	inline void deleteRootsAtPos (int i) {
		_root_size -= ((char *) _rootset[i].end - (char *) _rootset[i].start);
    _rootset[i].start = _rootset[_n_roots-1].start;
    _rootset[i].end = _rootset[_n_roots-1].end;
    _rootset[i].tmp = _rootset[_n_roots-1].tmp;
    _n_roots--;
  }
  
  /// Return the root at a given position in the root set, return an empty
  /// root object if the request is out of range.
  inline GC_root rootAtPos (int n) {
  	assert (n >= 0);
  	
  	if (n >= _n_roots) {
  		GC_root gr;
  		gr.start = 0;
  		gr.end = 0;
  		
  		return gr;
  	}
  	
  	return _rootset[n];
  }
  
  /// Clear out temporary roots, and re-register the root set to keep it up
	/// to date.
	void registerDynamicLibraries() {
		deleteTempRoots();
		
    if (registerDynlib_dl_iterate_phdr()) {
    	if (GC_VERBOSE && GC_VERBOSE_ROOTS) {
    		printRoots();
    	}
    	
			return;
    }
    
    char buf[255];
    sprintf (buf, "*** GC Unable to register dynamic libraries!\n");
    fprintf (stderr, buf);
  }
	
	/// We try to parse /self/maps for the dynamic roots, except it doesn't quite
	/// work as expected.  There's an alternative method below, though...
	void registerMapEntries (void) {
		deleteTempRoots();
		
		ByteWrapper maps = GCPlatform::gcReadMaps();
		assert (maps.ptr != 0);
		
		char *bumpPtr = maps.ptr;
		char prot_buf[5];
		gc_word start = 0;
		gc_word end = 0;
		unsigned int maj_dev = 0;
	
		while (1) {
			bumpPtr = GCPlatform::parseMapEntry (bumpPtr, &start, &end, prot_buf, &maj_dev);

			if (bumpPtr == 0) {
				MmapWrapper::unmap (maps.ptr, maps.size * sizeof(char));
				return;
			}
			
			if (prot_buf[1] == 'w') {
				// Check for and discard the stack range
				//if (start <= (word)GC_stackbottom && end >= (word)GC_stackbottom) {
				//	continue;
				//}
				
				// There should also be some code for the main data segment, here and
				// above.  We'll get to that...
				
				//addRoots ((char *) start, (char *) end, true);
			}
		}
	}
  
  /// Callback function to add roots to the root set.
  /// Taken (almost) verbatim from Boehm and I don't know how it works.
  int registerDynlibCallback(dl_phdr_info * info, size_t size) {

	  // Make sure struct dl_phdr_info is at least as big as we need.
	  if (size < offsetof (dl_phdr_info, dlpi_phnum) + sizeof (info->dlpi_phnum)) {
	    return -1;
	  }
	
	  const ElfW(Phdr) * p = info->dlpi_phdr;
	  for (int i = 0; i < (int)(info->dlpi_phnum); ((i++),(p++)) ) {
	    switch( p->p_type ) {
	    case PT_LOAD:
			{
		  	if( !(p->p_flags & PF_W) ) break;
		  	char * start = ((char *)(p->p_vaddr)) + info->dlpi_addr;
		  	addRoots(start, start + p->p_memsz, true);
			}
	    break;
	    default:
			break;
	    }
	  }

	  return 0;
	}
	
	/// Return the size of all the registered roots combined
	size_t rootSetSize (void) {
		size_t size = 0;
		
		for (int i = 0; i < _n_roots; i++) {
			size += (char *) _rootset[i].end - (char *)_rootset[i].start;
		}
		
		return size;
	}
	
	/// Good for debugging
	void printRoots (void) {
    size_t total = 0;
    char buf[255];
    
    sprintf (buf, "*** GC Root Set:\n");
    fprintf (stderr, buf);
    
    for (int i = 0; i < _n_roots; i++) {
        sprintf (buf, "\tFrom 0x%lx to 0x%lx [SZ: %i]", _rootset[i].start, _rootset[i].end,
        				 (char *)_rootset[i].end - (char *)_rootset[i].start);
        fprintf (stderr, buf);
        
        if (_rootset[i].tmp) {
            sprintf (buf, " (temporary)\n");
            fprintf (stderr, buf);
        } else {
            sprintf (buf, "\n");
            fprintf (stderr, buf);
        }
        
        total += (gc_word) _rootset[i].end - (gc_word) _rootset[i].start;
    }
    
    sprintf (buf, "\tTotal size: %ld\n", total);
    fprintf (stderr, buf);
    
    if (_root_size != total) {
     	sprintf (buf, "_root_size incorrect: %ld (expected: %ld)!!\n", _root_size, total);
    }
  }
	
private:
	
	/// An alternative method for capturing the root set on modern linux systems.
	bool registerDynlib_dl_iterate_phdr (void) {
  	if (dl_iterate_phdr) {
	    int did_something = 0;
	    dl_iterate_phdr(GCRootSet::registerDynLibDispatch, this);
	    
	    if (!did_something) {
				// dl_iterate_phdr may forget the static data segment in
				// statically linked executables.
				//GC_add_roots_inner(DATASTART, (char *)(DATAEND), TRUE);
		
				// #       if defined(DATASTART2)
	      // GC_add_roots_inner(DATASTART2, (char *)(DATAEND2), TRUE);
				// #       endif
	    }
	
	    return true;
  	} else {
   		return false;
 	 }
	}
	
	/// Dispatch to bind the instantiated callback to the register routine
	static int registerDynLibDispatch (dl_phdr_info * info, size_t size, void * ptr) {
		GCRootSet * disp = (GCRootSet *) ptr;
		int ret = disp->registerDynlibCallback (info, size);
		
		return ret;
	}

	DynamicArray<GC_root> _rootset;
	
	int	_n_roots;
	size_t _root_size;
};

#endif
