// -*- C++ -*-

#ifndef _GCHEAP_H_
#define _GCHEAP_H_

#include <signal.h>
#include <execinfo.h>
#include <assert.h>

#include "gcstack.h"
#include "gcmark.h"
#include "gccfg.h"
#include "gccontrol.h"

void term_handler (int signum, siginfo_t * si, void * v) {
	char buf[255];
	sprintf (buf, "CAUGHT SIGSEGV!\nFaulting Address: %08X\n", si->si_addr);
	fprintf (stderr, buf);

	exit (1);
}

template <class SuperHeap>
class GCHeap : public SuperHeap {

public:

	GCHeap () 
		: _gcStack (GCStack::getStack()),
			_mark (this),
			_bytes_allocated (0),
			_root_size (0)
	{
		if (GC_DISABLE_RANDOM) {
			srand (GC_RANDOM_SEED1);
		}
		
		if (GC_VERBOSE) {
			char buf[255];
			sprintf (buf, "*** Initializing DieHard + GC\n");
			fprintf (stderr, buf);
		}
		
		register_term_handler();
	}
	
	~GCHeap () { }
	
	/// Possibly helpful in debugging.  But probably not.
	void * register_term_handler () {
		struct sigaction new_action;
		new_action.sa_sigaction = term_handler;
		sigemptyset (&new_action.sa_mask);
		new_action.sa_flags = SA_SIGINFO;
		
		sigaction (SIGSEGV, &new_action, 0);		
	}

	inline void * malloc (size_t sz) {
		assert (GCControl::triggerCollect == 0);
		assert (GCControl::triggerAlways == 0);

		// The Boehm trigger model tests its invariant during allocation
		if (GC_COLLECT_MODE == 2) {
			int liveSize = SuperHeap::gcLiveSize();
			int heapSize = SuperHeap::gcHeapSize();
			int freeSize = heapSize - liveSize;
			
			if (_root_size == 0) {
				_root_size = _mark.rootSetSize();
			}
			
			int work = _root_size + liveSize;
			int rel_work = (work >> 2) * 2;
			int rel_bytes = (_bytes_allocated >> 2) * 3;

			if (rel_bytes > rel_work ) {
				GCControl::triggerCollect = 1;
				_root_size = 0;
			}
		}
		
		void * ptr = 0;
		
		if (GCControl::triggerCollect == 0) {
			ptr = SuperHeap::malloc (sz);
		}
		
		// If a collection has been triggered, we perform it now then re-allocate.
		if (GCControl::triggerCollect || GCControl::triggerAlways) {
			doCollection();
			ptr = SuperHeap::malloc (sz);
			assert (GCControl::triggerCollect == 0);
			GCControl::triggerAlways = 0;
		}
		
		if (GC_LOG_MALLOC) {
			char buf[255];
			sprintf (buf, "M: %08x (%i)\n", ptr, sz);
			fprintf (stderr, buf);
		}
		
		if (GC_VERBOSE && GC_WATCH_CLASS) {
			void * baddr = SuperHeap::gcResolve (GC_WATCH_CLASS, GC_WATCH_HEAP, GC_WATCH_INDEX);
			if ((char *)ptr == (char *)baddr) {
				char buf[255];
				sprintf (buf, "WATCH: MALLOC(%i) -> %08x\n", sz, ptr);
				fprintf (stderr, buf);
			}
		}
		
		if (ptr != 0) {
			_bytes_allocated += getSize (ptr);
		}
		
		return ptr;
	}
	
	/// Cut off frees (we should do this upstream, but it's transparent here)
	inline bool free (void * ptr) {
		if (GC_USING_FREE) {
			if (GC_VERBOSE && GC_LOG_MALLOC) {
				char buf[255];
				sprintf (buf, "F: %08x\n", ptr);
				fprintf (stderr, buf);
			}
			
			if (GC_VERBOSE && GC_WATCH_CLASS) {
				void * baddr = SuperHeap::gcResolve (GC_WATCH_CLASS, GC_WATCH_HEAP, GC_WATCH_INDEX);
				if ((char *)ptr == (char *)baddr) {
					char buf[255];
					sprintf (buf, "WATCH: FREE(%08x)\n", ptr);
					fprintf (stderr, buf);
				}
			}

			return SuperHeap::free (ptr);
		}
		
		return true;
	}
	
	/// Get the size of an object
	inline size_t getSize (void * ptr) {
		size_t size = SuperHeap::getSize (ptr);
		return size;
	}
	
	/// Searching for watched items in the heap
	inline void * resolve () {
		return SuperHeap::gcResolve (GC_WATCH_CLASS, GC_WATCH_HEAP, GC_WATCH_INDEX);
	}
	
	/// (Try to) Mark a pointer on the heap
	inline int markScan (void * ptr) {
		int mb = SuperHeap::markScan (ptr);
		return mb;
	}
	
	/// Performs a full heap scan
	inline void fullScan (void) {
		SuperHeap::fullScan();
	}
	
	/// Do a complete garbage collection cycle
	void doCollection (void) {
		char buf[255];
		if (GC_VERBOSE) {
			sprintf (buf, "*** GC Cycle Start (Trigger on: %i)\n", GCControl::triggerSizeClass);
			fprintf (stderr, buf);
		}
		
		// Clear out all the bitmaps
		SuperHeap::gcZeroBitmaps();
		
		// Mark the root set, including stack, static data region,
		// and static libraries (maybe?)
		_mark.push_roots();
		
		if (GC_HEAP_FULLSCAN) {
			SuperHeap::fullScan();
		}
		
		// Scan the heap, using pointers from the root set.
		_mark.markRemaining();

		if (GC_USING_FREE) {
			SuperHeap::gcBitmapCheck();
		}
		
		// LargeHeap must be swept
		SuperHeap::gcSweep();
		
		// Force free memory to a "dead" value
		if (GC_BURY_DEAD) {
			SuperHeap::gcBuryDead();
		}
		
		// Optionally print a specific object in the heap each collection
		if (GC_WATCH_CLASS && GC_WATCH_RANGE) {
			void * baddr = SuperHeap::gcResolve (GC_WATCH_CLASS, GC_WATCH_HEAP, GC_WATCH_INDEX);
			_mark.printRange (baddr, (char *) baddr + GC_WATCH_CLASS);
		}
		
		// Try growing the heap, given the current size of the root set.
		SuperHeap::gcGrow (_mark.rootSetSize(), _bytes_allocated );
		
		if (GC_VERBOSE) {
			sprintf (buf, "Greatest Plausible Heap Address:\t%08x\nLeast Plausible Heap Address:\t%08x\n",
							 GCControl::greatest_plausible_heap_addr, GCControl::least_plausible_heap_addr);
			fprintf (stderr, buf);
			sprintf (buf, "*** GC Cycle Complete\n");
			fprintf (stderr, buf);
		}
		
		GCControl::triggerCollect = 0;
		GCControl::triggerSizeClass = 0;
		_bytes_allocated = 0;
	}
	
private:
	
	GCMark<GCHeap> _mark;
	GCStack * _gcStack;
	
	int _root_size;
	int _bytes_allocated;
};

#endif
