// -*- C++ -*-

#ifndef _GC_MARK_
#define _GC_MARK_

#include <setjmp.h>

#include "dynstack.h"
#include "gcstack.h"
#include "gcplatform.h"
#include "gcdynamic.h"
#include "gccontrol.h"

template <class Allocator>
class GCMark {

public:
	
	GCMark (Allocator * alloc)
		: _markStack (),
			_gcStack (GCStack::getStack()),
			_stackBottom (GCPlatform::gcStackBase()),
			_dataStart (GCPlatform::gcDataStart()),
			_dataEnd (GCPlatform::gcDataEnd()),
			_alloc (alloc)
		{ }
	
	/// Scan objects on the mark stack for new pointers until the mark stack
	/// is exhausted.
	inline void markRemaining (void) {
		if (GC_VERBOSE && GC_VERBOSE_TRACE) {
			char buf[255];
			sprintf (buf, "*** Mark recursively from MARK STACK\n");
			fprintf (stderr, buf);
		}
		
		while (!_gcStack->empty()) {
			GC_mse mse = _gcStack->pop();
			markRange (mse);
		}
	}
	
	/// Scan the stack (and registers) for live objects
	inline void markStack () {
		if (GC_VERBOSE && GC_VERBOSE_TRACE) {
			char buf[255];
			sprintf (buf, "*** Mark REGISTERS and STACK\n");
			fprintf (stderr, buf);
		}

		__builtin_unwind_init();

		if (GC_VERBOSE_STACK) {
			GCControl::verboseRange = 1;
		}
		
		// Grab the current stack pointer and work backwards
		int stackTop = 0;

		markRange (&stackTop, _stackBottom);
	}
	
	/// Scan dynamic libraries (and other regions registered as the root set)
	/// for pointers into the heap.
	inline void markDynamic () {
		if (GC_VERBOSE && GC_VERBOSE_TRACE) {
			char buf[255];
			sprintf (buf, "*** Mark DYNAMIC SHARED MEMORY\n");
			fprintf (stderr, buf);
		}
		
		_grs.registerDynamicLibraries();
		
		// Scan each registered root
		for (int i=0; ; i++) {
			GC_root gr = _grs.rootAtPos (i);
			if (gr.start == 0 || gr.end == 0) {
				return;
			}
			
			markRange (gr.start, gr.end);
		}
	}
	
	/// Scan the static data region for pointers.  This may already be handled
	/// by markDynamic.
	inline void markData () {
		if (GC_VERBOSE && GC_VERBOSE_TRACE) {
			char buf[255];
			sprintf (buf, "*** Mark DATA SEGMENT\n");
			fprintf (stderr, buf);
		}
		
		markRange (_dataStart, _dataEnd);
	}
	
	/// Mark a range of memory, given a memory region object (mse)
	inline void markRange (GC_mse mse) {
		void * top = mse.address;
		void * bottom = (void *)((gc_word)top + mse.length);
		
		markRange (top, bottom);
	}
	
	/// Scan 4-byte-aligned block of memory in a given range, looking for
	/// live pointers.
	inline void markRange (void * bottom, void * top) {

		// Convert each block of memory to a potential pointer, and attempt to mark
		// it.
		for (void * ptr = (void*)((gc_word)top-4);
			   ptr >= bottom; 
			   ptr = (void *)((gc_word)ptr - 4))
		{
			void * valPtr = (void *)(*(gc_word*)ptr);
			
			if (GCControl::verboseRange) {
				char buf[50];
				sprintf (buf, "  %08x (%08x)\n", ptr, valPtr);
				fprintf (stderr, buf);
			}
			
			if (GC_VERBOSE && GC_WATCH_CLASS) {
				void * baddr = _alloc->resolve();
				if (baddr == valPtr && baddr > 0) {
					char buf[255];
					sprintf (buf, "WATCH: SCAN(%08x) [AT: %08x, OBJ: %08x, SZ: %i]\n", valPtr, ptr, bottom, (char*)top - (char*)bottom);
					fprintf (stderr, buf);
				}
			}

			markPointer (valPtr);
		}
		
		GCControl::verboseRange = 0;
	}
	
	inline void printRange (void * bottom, void * top) {
		char buf[50];
		sprintf (buf, "*** PRINT RANGE: %08x - %08x (%i)\n", bottom, top, (char* )top - (char* )bottom);
		fprintf (stderr, buf);
		
		for (void * ptr = (void*)((gc_word)bottom);
			   ptr < top; 
			   ptr = (void *)((gc_word)ptr + 4))
		{
			void * valPtr = (void *)(*(gc_word*)ptr);
			
			sprintf (buf, "  %08x (%08x)\n", ptr, valPtr);
			fprintf (stderr, buf);
		}
	}
	
	/// Attempt to mark a pointer
	inline void markPointer (void * ptr) {

		// Is the pointer out of mappable range?
		if ((ptr_t) ptr >= (ptr_t) GCControl::greatest_plausible_heap_addr ||
			  (ptr_t) ptr < (ptr_t) GCControl::least_plausible_heap_addr) {
			return;
		}
		
		// Call on GCHeap in the allocator hierarchy to try marking a pointer in
		// the lower layers of the heap.  
		int mb = _alloc->markScan (ptr);
		
		// Note, mb no longer serves a purpose.
	}
	
	/// Calculate the size of the root set, including stack, static regions,
	/// dynamic libraries, etc.
	inline size_t rootSetSize (void) {
		void * stackTop;
		stackTop = &stackTop;
		
		int roots1 = (char *)_stackBottom - (char *)stackTop;
		int roots2 = (char *)_dataEnd - (char *)_dataStart;
		
		if (roots1 < 0) roots1 *= -1;
		if (roots2 < 0) roots2 *= -1;
		
		return roots1 + roots2 + _grs.rootSetSize();
	}
	
	/// The below routines are an alternate method of scanning that are a direct
	/// equivilant to Boehm's scanning code
	
	void mark_from (void) {
		if (GC_VERBOSE && GC_VERBOSE_TRACE) {
			char buf[255];
			sprintf (buf, "*** Mark recursively from MARK STACK\n");
			fprintf (stderr, buf);
		}
		
		while (!_gcStack->empty()) {
			GC_mse mse = _gcStack->pop();
			push_all_eager ((ptr_t) mse.address, (ptr_t) mse.address + mse.length);
		}
	}
	
	inline void mark_and_push_stack (ptr_t p) {
		size_t sz = _alloc->getSize (p);
		if (sz > 0) {
			size_t objSz = _alloc->markScan (p);
			if (objSz > -1) {
				size_t diff = objSz - sz;
				push_all (p - diff, p + objSz);
			}
		}
	}
	
	void push_all (ptr_t bottom, ptr_t top) {
		bottom = (ptr_t)(((gc_word) bottom + GC_ALIGNMENT-1) & ~(GC_ALIGNMENT-1));
		top = (ptr_t)(((gc_word) top) & ~(GC_ALIGNMENT-1));
		if (top == 0 || bottom == top) {
    	return;
    }
    
    GC_mse mse;
    mse.address = bottom;
    mse.length = top - bottom;
    
    _gcStack->push (mse);
	}
	
	void push_all_eager (ptr_t bottom, ptr_t top) {
		gc_word * b = (gc_word *)(((gc_word) bottom + GC_ALIGNMENT-1) & ~(GC_ALIGNMENT-1));
    gc_word * t = (gc_word *)(((gc_word) top) & ~(GC_ALIGNMENT-1));
    if (top == 0) {
    	return;
    }

    gc_word * lim = t - 1;
    for (gc_word * p = b; p <= lim; p = (gc_word *)(((char *)p) + GC_ALIGNMENT)) {
			gc_word q = *p;
			if ((ptr_t) (q) >= (ptr_t) GCControl::least_plausible_heap_addr	&& 
					(ptr_t) (q) < (ptr_t) GCControl::greatest_plausible_heap_addr) {
				
				if (GC_VERBOSE && GC_WATCH_CLASS) {
					void * baddr = _alloc->resolve();
					if (baddr == (ptr_t)(q) && baddr > 0) {
						char buf[255];
						sprintf (buf, "WATCH: SCAN(%08x) [AT: %08x, OBJ: %08x, SZ: %i]\n", q, p, bottom, (char*)top - (char*)bottom);
						fprintf (stderr, buf);
					}
				}
			
				mark_and_push_stack ((ptr_t) (q));
			}
    }
	}
	
	ptr_t approx_sp () {
    volatile gc_word dummy = 42;
				/* Force stack to grow if necessary.	Otherwise the	*/
    		/* later accesses might cause the kernel to think we're	*/
    		/* doing something wrong.				*/
    ptr_t dummy2 = (ptr_t) &dummy;
    return dummy2;
  }

	void push_current_stack (void) {
		ptr_t stTop = approx_sp();
		if (GC_VERBOSE && GC_VERBOSE_ROOTS) {
			char buf[255];
			sprintf (buf, "\tFrom 0x%lx to 0x%lx [SZ: %i]\n", _stackBottom, stTop,
        			(char *)_stackBottom - (char*)stTop);
      fprintf (stderr, buf);
    }
    
		push_all_eager (stTop, (ptr_t) _stackBottom);
	}

	void noop(gc_word x) {
    static volatile gc_word sink = x;
  }
  
	void generic_push_regs (void) {
		gc_word dummy;
		
		__builtin_unwind_init();
		push_current_stack();
		
		noop ((gc_word) &dummy);
	}
	
	void push_roots (void) {
		if (GC_VERBOSE && GC_VERBOSE_TRACE) {
			char buf[255];
			sprintf (buf, "*** Mark DYNAMIC SHARED MEMORY\n");
			fprintf (stderr, buf);
		}
		
		_grs.registerDynamicLibraries();

		// Scan each registered root
		for (int i=0; ; i++) {
			GC_root gr = _grs.rootAtPos (i);
			if (gr.start == 0 || gr.end == 0) {
				break;
			}
			
			push_all_eager ((ptr_t) gr.start, (ptr_t) gr.end);
		}
		
		if (GC_VERBOSE && GC_VERBOSE_TRACE) {
			char buf[255];
			sprintf (buf, "*** Mark DATA SEGMENT\n");
			fprintf (stderr, buf);
		}
		
		push_all_eager ((ptr_t) _dataStart, (ptr_t) _dataEnd);
		if (GC_VERBOSE && GC_VERBOSE_ROOTS) {
			char buf[255];
			sprintf (buf, "\tFrom 0x%lx to 0x%lx [SZ: %i]\n", _dataStart, _dataEnd,
        			(char *)_dataEnd - (char*)_dataStart);
      fprintf (stderr, buf);
    }
		
		if (GC_VERBOSE && GC_VERBOSE_TRACE) {
			char buf[255];
			sprintf (buf, "*** Mark REGISTERS and STACK\n");
			fprintf (stderr, buf);
		}
		
		generic_push_regs();
	}
	
private:
	
	void * _stackBottom;
	void * _dataStart;
	void * _dataEnd;
	
	Allocator * _alloc;
	GCStack * _gcStack;
	GCRootSet _grs;
	DynamicStack<void *> _markStack;
	
	
	/*inline push_one_stack (p, source) {
		if ((ptr_t)(p) >= (ptr_t)GC_least_plausible_heap_addr	&& 
				(ptr_t)(p) < (ptr_t)GC_greatest_plausible_heap_addr) {
	 	PUSH_ONE_CHECKED_STACK(p, source);
	}
	
	inline push_one_checked_stack (p, source) {
		GC_mark_and_push_stack(p);
	}*/
};

#endif
