// -*- C++ -*-

#ifndef _GCMINIHEAP_H_
#define _GCMINIHEAP_H_

#include <assert.h>

#include "../randomminiheap.h"
#include "dynarray.h"
#include "gcstack.h"
#include "gccfg.h"
#include "gccontrol.h"

struct bitmapLocation {
	int	sizeClass;
	int	index;
};

template <int Numerator,
	  int Denominator,
	  unsigned long ObjectSize,
	  unsigned long NObjects,
	  class Allocator,
	  bool DieFastOn>
class GCMiniHeap : public RandomMiniHeap<Numerator, Denominator, ObjectSize, NObjects, Allocator, DieFastOn> {

	typedef RandomMiniHeap<Numerator, Denominator, ObjectSize, NObjects, Allocator, DieFastOn> MiniHeap;

	using MiniHeap::getSize;
	using MiniHeap::getObject;
	using MiniHeap::computeIndex;
	using MiniHeap::_miniHeapBitmap;
	using MiniHeap::_random;
	using MiniHeap::_miniHeap;
	
public:
	
	GCMiniHeap (void)
		: _gcStack (GCStack::getStack())
		{ }
	
	inline void * malloc (size_t sz) {
		int prevInUse = _miniHeapBitmap.active();
  	
  	void * mcr = MiniHeap::malloc (sz);
  	
  	if ((int *)mcr == NULL) {
 			return NULL;
		}
		 	
  	if (GC_USING_FREE) {
  		int index = (int) modulo<NObjects> (_random.last());
  		if (!_markBitmap.tryToSet (index) ) {
  			_miniHeapBitmap.reset (index);
 			
  			assert (prevInUse == _miniHeapBitmap.active() );
  			return NULL;
  		}
  	}
  	
  	assert (prevInUse + 1 == _miniHeapBitmap.active() );
		return mcr;
	}
	
	inline bool free (void * ptr) {
  	bool ret = MiniHeap::free (ptr);
  	
  	// If free failed, this code would cause horrible things to happen.  Why?
  	if (ret) {
	  	int index = computeIndex (ptr);
   		assert ((index >= 0) && (index < NObjects));
  		assert (!_miniHeapBitmap.isSet (index) );
  	}
    
    return ret;
	}
	
	/// Extend the miniheap activate to initialize the markmap
	inline void activate (void) {
		MiniHeap::activate();		
		//_gcStack = GCStack::getStack();
		
		if (GC_USING_FREE) {
			_markBitmap.reserve (NObjects);
			_markBitmap.clear();
		}
		
		// Maintain lower and upper bounds on heap locations
		if ((ptr_t) _miniHeap <= (ptr_t) GCControl::least_plausible_heap_addr ||
        (ptr_t) GCControl::least_plausible_heap_addr == 0) {
    	GCControl::least_plausible_heap_addr = ((ptr_t) _miniHeap - sizeof(gc_word));
    }
    
    gc_word bytes = ObjectSize * NObjects;
    
    if ((ptr_t) _miniHeap + bytes >= (ptr_t) GCControl::greatest_plausible_heap_addr) {
    	GCControl::greatest_plausible_heap_addr = (ptr_t) _miniHeap + bytes;
    }
	}
	
	/// Clear the internal bitmap, which doubles as a markmap.
	inline void zeroBitmap (void) {
		if (GC_USING_FREE) {
			if ((GC_EXCLUDE_HEAPS | ObjectSize) == GC_EXCLUDE_HEAPS) {
				_markBitmap.clear();
				for (int index = 0; index < NObjects; index++) {
					if (_miniHeapBitmap.isSet (index) ) {
						markScan (getObject (index) );
					}
				}
			}
			_markBitmap.clear();
		} else {
			_miniHeapBitmap.clear();
		}
	}
	
	/// When USING_FREE is enabled, compares number of undeleted objects
	/// to number of reclaimed objects, reports differences.
	void bitmapCheck (void) {
		int early = 0;
		int late = 0;
		int gc = 0;
		int mf = 0;
		
		DynamicArray<bitmapLocation> bmp;

		for (int index = 0; index < NObjects; index++) {
			if (_markBitmap.isSet (index) && !_miniHeapBitmap.isSet (index) ) {
				late++;
			}
			if (!_markBitmap.isSet (index) && _miniHeapBitmap.isSet (index) ) {
				bitmapLocation b;
				b.index = index;
				bmp[early] = b;
				
				early++;
			}
			if (_markBitmap.isSet (index) ) {
				gc++;
			}
			if (_miniHeapBitmap.isSet (index) ) {
				mf++;
			}
		}
		
		if (GC_VERBOSE && GC_VERBOSE_BITMAPS) {
			char buf[255];
			sprintf (buf, "    Objs: %6i  Early: %6i  Late: %6i  GC: %4i/%4i  MF: %4i/%4i\n", NObjects, early, late, _markBitmap.active(), gc, _miniHeapBitmap.active(), mf);
			fprintf (stderr, buf);
			
			if (early > 0) {
				sprintf (buf, "      Early: ");
				fprintf (stderr, buf);
				
				for (int x = 0; x < early; x++) {
					sprintf (buf, "%i (%08x), ", bmp[x].index, getObject (bmp[x].index) );
					fprintf (stderr, buf);
				}
				
				sprintf (buf, "\n");
				fprintf (stderr, buf);
			}
		}
		
		if (GC_ASSERT_BITMAPS) {
			_miniHeapBitmap.count();
			_markBitmap.count();
		}
		
		if (GC_ENFORCING_FREE || (GC_EXCLUDE_HEAPS | ObjectSize) == GC_EXCLUDE_HEAPS) {
			_markBitmap.copy (_miniHeapBitmap);
		} else {
			//_miniHeapBitmap.copy (_markBitmap);
			_miniHeapBitmap.intersect (_markBitmap);
		}
	}
	
	/// Walk through the heap and set cleared objects to a DEAD value
	inline void buryDead (void) {
		for (int index = 0; index < NObjects; index++) {
			if ( (GC_USING_FREE && !_markBitmap.isSet (index) ) ||
					 (!_miniHeapBitmap.isSet (index) ) ) 
			{
				size_t * object = (size_t *) getObject (index);
				
				for (int x = 0; x < ObjectSize; x += 4) {
					*object = GC_BURY_VALUE;
					object++;
				}
			}
		}
	}
	
	inline void * resolve (size_t gcindex) {
  	return getObject (gcindex);
  }
	
	inline int markScan (void * ptr) {

  	// Get size to check in object, doubles as a bounds check
  	size_t szRemaining = getSize (ptr);

  	if (szRemaining == 0) {
  		return 0;
  	}

  	// Ensure that the mark bit for this object is set
  	int index = computeIndex (ptr);
    assert ((index >= 0) && (index < NObjects));
    
    // How far "into" an object are we pointing?
    size_t diff = ObjectSize - szRemaining;

		// Check if this object is already marked
		if (GC_USING_FREE) {
			if (!_markBitmap.tryToSet (index) ) {
				return -1;
			}
		} else {
	    if (!_miniHeapBitmap.tryToSet (index) ) {
  	  	return -1;
  	  }
  	}
    
    // Push start of object to mark stack
    GC_mse entry;
    entry.address = (void *)((char *) ptr - diff);
    entry.length = ObjectSize;
    
    _gcStack->push (entry);
  	
  	// I don't think markScan signalling is used anymore.
  	return ObjectSize;
  }
  
  inline void fullScan (void) {
  	GC_mse entry;
  	entry.address = _miniHeap;
  	entry.length = NObjects * ObjectSize;
  	_gcStack->push (entry);
  }
	
private:
	
	GCStack* _gcStack;
	
	BitMap<Allocator> _markBitmap;
	
};

#endif
