// -*- C++ -*-

#ifndef _GC_PLATFORM_
#define _GC_PLATFORM_

#include <assert.h>
#include <setjmp.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/fcntl.h>

#include "../mmapwrapper.h"

#pragma weak __data_start
extern int __data_start[];

#pragma weak data_start
extern int data_start[];
    
//extern int _etext[];
extern int _end[];

typedef unsigned long gc_word;
typedef long gc_signed_word;
typedef char * ptr_t;

typedef void (*handler)(int);

struct ByteWrapper {
	char * ptr;
	size_t size;
};

enum { GC_ALIGNMENT = 4 };

/// Very platform-specific routines for Linux/x86
class GCPlatform {

enum { STAT_SKIP = 27 };
enum { STAT_BUF_SIZE = 4096 };

public:
	
	static void * gcStackBase (void) {
		// We read the stack base value from /proc/self/stat.  We do this
    // using direct I/O system calls in order to avoid calling malloc
    // in case REDIRECT_MALLOC is defined.
    
		char stat_buf[STAT_BUF_SIZE];

    int f = open ("/proc/self/stat", O_RDONLY);
    if (f < 0 || read (f, stat_buf, STAT_BUF_SIZE) < 2 * STAT_SKIP) {
    	fprintf (stderr, "Couldn't read /proc/self/stat\n");
    	exit (1);
    }
    
    size_t buf_offset = 0;
    char c = stat_buf[buf_offset++];
    
    // Skip the required number of fields.  This number is hopefully
    // constant across all Linux implementations.
    for (size_t i = 0; i < STAT_SKIP; ++i) {
			while (isspace (c)) {
				c = stat_buf[buf_offset++];
			}
			while (!isspace (c)) {
				c = stat_buf[buf_offset++];
			}
    }
    
    while (isspace (c)) {
    	c = stat_buf[buf_offset++];
    }
    
    gc_word result = 0;
    while (isdigit (c)) {
      result *= 10;
      result += c - '0';
      c = stat_buf[buf_offset++];
    }
    
    close (f);
    if (result < 0x10000000) {
    	fprintf (stderr, "Absurd stack bottom value");
    	exit (1);
    }
    
    char buf[255];
    sprintf (buf, "*** GC Stack Base: %x\n", result);
    fprintf (stderr, buf);
    
    return (void *) result;
	}
	
	/// Parse self/maps for getting dynamic roots, not used in second dynamic
	/// root alternative.
	static inline ByteWrapper gcReadMaps (void) {
    size_t maps_size = 4096;
    size_t maps_buf_sz = 2048;
    char * maps_buf = 0;
    
    ByteWrapper bw;
    bw.ptr = 0;
    bw.size = 0;
    
    do {
    	if (maps_size >= maps_buf_sz) {
    		if (maps_buf != 0) {
    			MmapWrapper::unmap (maps_buf, maps_buf_sz * sizeof(char));
    		}
    		
    		while (maps_size >= maps_buf_sz) {
    			maps_buf_sz *= 2;
    		}
    	
	    	maps_buf = (char *) MmapWrapper::map (maps_buf_sz * sizeof(char));
	    	if (maps_buf == 0) {
	    		return bw;
	    	}
	    }
    	
    	int f = open("/proc/self/maps", O_RDONLY);
	    if (f == -1) {
	    	return bw;
	    }
	    
	    maps_size = 0;
	    int result = 0;
	    do {
        result = GCPlatform::repeat_read (f, maps_buf, maps_buf_sz-1);
        if (result <= 0) {
        	return bw;
        }
        
        maps_size += result;
	    } while (result == maps_buf_sz-1);
	    
	    close(f);
	  } while (maps_size >= maps_buf_sz);
	  
	  maps_buf[maps_size] = '\0';
	  
	  char buf[maps_size+64];
	  sprintf (buf, "%s", maps_buf);
	  fprintf (stderr, buf);
		
		bw.ptr = maps_buf;
		bw.size = maps_size;
		
		return bw;
	}
	
	//  GC_parse_map_entry parses an entry from /proc/self/maps so we can
	//  locate all writable data segments that belong to shared libraries.
	//  The format of one of these entries and the fields we care about
	//  is as follows:
	//  XXXXXXXX-XXXXXXXX r-xp 00000000 30:05 260537     name of mapping...\n
	//  ^^^^^^^^ ^^^^^^^^ ^^^^          ^^
	//  start    end      prot          maj_dev
	//
	//  Note that since about auguat 2003 kernels, the columns no longer have
	//  fixed offsets on 64-bit kernels.  Hence we no longer rely on fixed offsets
	//  anywhere, which is safer anyway.
	//
	
	// Assign various fields of the first line in buf_ptr to *start, *end,
	// *prot_buf and *maj_dev.  Only *prot_buf may be set for unwritable maps.

	static inline char * parseMapEntry (char * buf_ptr, gc_word * start, gc_word * end,
  										  			 char * prot_buf, unsigned int * maj_dev) {
    if (buf_ptr == NULL || * buf_ptr == '\0') {
        return 0;
    }

    char * p = buf_ptr;
    while (isspace (*p)) p++;
    
    char * start_start = p;
    char * endp;
    assert (isxdigit (*start_start));
    *start = strtoul (start_start, &endp, 16);
    p = endp;
    assert (*p == '-');

    p++;
    char * end_start = p;
    assert (isxdigit (*end_start));
    *end = strtoul (end_start, &endp, 16);
    p = endp;
    assert (isspace (*p));

    while (isspace (*p)) p++;
    
    char * prot_start = p;
    assert (*prot_start == 'r' || *prot_start == '-');
    memcpy (prot_buf, prot_start, 4);
    prot_buf[4] = '\0';
    
    if (prot_buf[1] == 'w') {
    	// we can skip the rest if it's not writable.
			// Skip past protection field to offset field
      while (!isspace (*p)) p++;
      while (isspace (*p)) p++;
      
      assert (isxdigit (*p));
      
			// Skip past offset field, which we ignore
      while (!isspace (*p)) p++;
      while (isspace (*p)) p++;
      
			char * maj_dev_start = p;
      assert (isxdigit (*maj_dev_start));
      *maj_dev = strtoul (maj_dev_start, 0, 16);
    }

    while (*p && *p++ != '\n');

    return p;
	}
	
	/// Get the start of the application's static data region.
	static inline void * gcDataStart (void) {
		//gc_word datastart = ((((gc_word) (_etext)) + 0xfff) & ~0xfff);
		ptr_t GC_data_start = 0;
		
		if ((ptr_t) __data_start != 0) {
	  	GC_data_start = (ptr_t)(__data_start);
    }   
    else if ((ptr_t) data_start != 0) {
	  	GC_data_start = (ptr_t)(data_start);
    }
    
    if (GC_data_start == 0) {
    	GC_data_start = GC_find_limit ((ptr_t) _end, false);
    }
    
		char buf[255];
		sprintf (buf, "*** GC Data Start: %x\n", GC_data_start);
		fprintf (stderr, buf);
		
		return (void *) GC_data_start;
		
		//return (void *)((((gc_word) (_etext)) + 0xfff) & ~0xfff);
	}
	
	/// Get the end of the application's static data region.
	static inline void * gcDataEnd (void) {
		gc_word dataend = (gc_word)_end;
		char buf[255];
		sprintf (buf, "*** GC Data End: %x\n", dataend);
		fprintf (stderr, buf);
		
		return (void *)_end;
	}
	
	/// Putting registers on the stack on an x86 platform
	static inline void gcPushRegs (void) {
		asm("pushl %ebp; addl $4,%esp");
	  asm("pushl %esi; addl $4,%esp");
	  asm("pushl %edi; addl $4,%esp");
	  asm("pushl %ebx; addl $4,%esp");
	}
	
	static size_t repeat_read(int fd, char * buf, size_t count) {
    size_t num_read = 0;

	  while (num_read < count) {
			size_t result = read(fd, buf + num_read, count - num_read);
			
			if (result < 0) {
				return result;
			}
			if (result == 0) {
				break;
			}
			
			num_read += result;
	  }
	  
	  return num_read;
	}
	
	// Locating mapped page boundaries
	
	
	
	enum { MIN_PAGE_SIZE = 256 };
	
	static void GC_set_and_save_fault_handler (handler h) {
		GC_old_segv_handler = signal (SIGSEGV, h);
	}
	
	static void GC_fault_handler (int sig) {
  	siglongjmp (GC_jmp_buf, 1);
  }
  
  static void GC_setup_temporary_fault_handler () {
		GC_set_and_save_fault_handler (GC_fault_handler);
  }
  
  static void GC_reset_fault_handler () {
  	signal (SIGSEGV, GC_old_segv_handler);
  }
  
  static void GC_noop (gc_word x) {
  	static volatile gc_word sink;
  	sink = x;
  }
  
  static ptr_t GC_find_limit (ptr_t p, bool up) {
  	static volatile ptr_t result;
  	
  	GC_setup_temporary_fault_handler();
		if (sigsetjmp (GC_jmp_buf, 1) == 0) {
	    result = (ptr_t)(((gc_word)(p)) & ~(MIN_PAGE_SIZE-1));
	    while (1) {
 	    	if (up) {
		    	result += MIN_PAGE_SIZE;
 	      } else {
		    	result -= MIN_PAGE_SIZE;
 	      }
 	      
				GC_noop ((gc_word)(*result));
	  	}
		}
		
		GC_reset_fault_handler();
 		if (!up) {
	    result += MIN_PAGE_SIZE;
 		}
 		
		return result;
  }

protected:
	
	static sigjmp_buf GC_jmp_buf;
	static handler GC_old_segv_handler;
};

handler GCPlatform::GC_old_segv_handler = 0;

sigjmp_buf GCPlatform::GC_jmp_buf;

#endif
