// -*- C++ -*-

#ifndef _GCRANDOMHEAP_H_
#define _GCRANDOMHEAP_H_

#include <assert.h>

#include "../randomheap.h"
#include "gccfg.h"
#include "gccontrol.h"

template <int Numerator,
		  int Denominator,
          size_t ObjectSize,
	  	  size_t AllocationGrain,
	  	  template <int Numerator,
		    		int Denominator,
		    		unsigned long ObjectSize,
		    		unsigned long NObjects,
		    		class Allocator,
		    		bool DieFastOn> class MiniHeap,
	  	  bool DieFastOn>
class GCRandomHeap : public RandomHeap<Numerator, Denominator, ObjectSize, AllocationGrain, MiniHeap, DieFastOn> {
	
	typedef RandomHeap<Numerator, Denominator, ObjectSize, AllocationGrain, MiniHeap, DieFastOn> ParentHeap;
	
	using ParentHeap::getMiniHeap;
	using ParentHeap::getAnotherMiniHeap;
	using ParentHeap::_miniHeapsInUse;
	using ParentHeap::_inUse;
	using ParentHeap::_available;

public:
	
	GCRandomHeap (void)
		: _markInUse (0)
	{	}
	
	/// Wraps parent malloc to override default growth behavior, and instead
  /// triggers a collection cycle.
  inline void * malloc (size_t sz) {
  	bool abortAlloc = false;
  	
		if (GC_ALWAYS_COLLECT) {
			if (GCControl::triggerAlways == 0) {
				GCControl::triggerAlways = 1;
				GCControl::triggerSizeClass = ObjectSize;
				abortAlloc = true;
			}
		}
		
		if (GC_COLLECT_MODE == 1) {
	  	if (_available > 0 && Numerator * _markInUse >= _available * Denominator) {
 	 			GCControl::triggerCollect = 1;
 	 			GCControl::triggerSizeClass = ObjectSize;
  			return 0;
  		}
  	}
  	
  	if (abortAlloc) {
  		return 0;
  	}

  	void * ptr = ParentHeap::malloc (sz);
  	
  	_markInUse++;

  	return ptr;
  }
  
  /// Wraps parent free for little apparent reason right now.
  inline bool free (void * ptr) {
    return ParentHeap::free (ptr);
  }
  
  /// Clear the bitmaps for each miniheap under this object
	inline void zeroBitmap (void) {
		for (int i = _miniHeapsInUse - 1; i >= 0; i--) {
      getMiniHeap(i)->zeroBitmap();
    }
    
    _markInUse = 0;
    if (!GC_USING_FREE) {
	    _inUse = 0;
	  }
  }
  
  /// When USING_FREE is enabled, check bitmaps in all miniheaps for
  /// inconsistencies
  inline void bitmapCheck (void) {
  	if (GC_VERBOSE && GC_VERBOSE_BITMAPS) {
	  	char buf[255];
 	 		sprintf (buf, "  Size Class: %i\n", ObjectSize);
 	 		fprintf (stderr, buf);
 	 	}
  	
  	for (int i = _miniHeapsInUse - 1; i >= 0; i--) {
  		getMiniHeap(i)->bitmapCheck();
  	}
  	
  	if (GC_ENFORCING_FREE) {
	  	_markInUse = _inUse;
	  } else {
	  	_inUse = _markInUse;
	  }
  }
  
  /// Force a default value into all the dead memory
  inline void buryDead (void) {
  	for (int i = _miniHeapsInUse - 1; i >= 0; i--) {
  		getMiniHeap(i)->buryDead();
  	}
  }
  
  inline void * resolve (size_t gcheap, size_t gcindex) {
  	if (gcheap >= _miniHeapsInUse) {
  		return 0;
  	}
  	
  	return getMiniHeap(gcheap)->resolve (gcindex);
  }
  
  /// Return the number of objects in use
  inline size_t objectsInUse (void) {
  	return _markInUse;
  }
  
  /// Return the object capacity of all miniheaps under this object
  inline size_t objectsAvailable (void) {
  	return _available;
  }
  
  /// Add a new miniheap to this object if it meets the growth condition
  inline void gcGrow (size_t liveSize) {
  	if (_available == 0) {
  		return;
  	}
  	
		// Strengthens growth condition, needed to safely use fractional C's
		// Hi/C > K + L + MCLi
		// This may have other unexpected implications!!
		liveSize += (_markInUse * ObjectSize * GC_DENOMINATOR * Numerator) / (GC_NUMERATOR * Denominator);

		// Grow heap if necessary, report any fatal out-of-space conditions if
		// growth fails where it's needed...
		if (GC_COLLECT_MODE == 1) {
	  	if (_available * ObjectSize * GC_DENOMINATOR < liveSize * GC_NUMERATOR) {
	  		if (GC_VERBOSE) {
	  			char buf[255];
					sprintf (buf, "*** GC Grow RandomHeap (%i)\n", ObjectSize);
					fprintf (stderr, buf);
				}
			
				getAnotherMiniHeap();
			}
			else if (Numerator * _markInUse >= _available * Denominator) {
				if (GC_VERBOSE) {
					char buf[255];
					sprintf (buf, "XXX GC Heap Starvation (%i)\n", ObjectSize);
					fprintf (stderr, buf);
				}
				
				getAnotherMiniHeap();
			}
		}
		else {
			// Instead we're using the Boehm model so we let DieHard do its own thing
			if (Numerator * _markInUse >= _available * Denominator) {
				if (GC_VERBOSE) {
					char buf[255];
					sprintf (buf, "XXX GC Heap Starvation [normal] (%i)\n", ObjectSize);
					fprintf (stderr, buf);
				}
				
				getAnotherMiniHeap();
			}
		}
  }
  
  /// Look for a miniheap containing the given pointer, add it back to the
  /// live set if found and not already marked.
  inline int markScan (void * ptr) {
		
		for (int i = _miniHeapsInUse - 1; i >= 0; i--) {
			int mb = getMiniHeap(i)->markScan (ptr);
			if (mb != 0) {
				if (mb > 0) {
			  	_markInUse += 1;
			  
			  	if (!GC_USING_FREE) {
						_inUse += 1;
					}
				}
				
				return mb;
			}
		}
		
		// not found in this heap
		return 0;
	}
	
	inline void fullScan (void) {
		for (int i = _miniHeapsInUse - 1; i >= 0; i--) {
			getMiniHeap(i)->fullScan ();
		}
	}

private:

	size_t _markInUse;

};

#endif
